/* 
        DebugMode.cc

        MASL/M-MASL Copyright (c) KUNI/TDU
        Windows version Copyright (c) 1995, 1996 Tomoatsu Shimada  (shimarin)
        X Window System version Copyright (c)1997 Yamahata Isaku (ISA-9)
*/
#include "DebugMode.h"
#ifdef DEBUG_MODE

const char DebugMode::arrow[]="->";

DebugMode::DebugMode (const char* const scenarioStart,
		      const int scenarioLength,
		      const int* const intVar,Bool isBreakMode=False,
		      const char* const maslTty=NULL):
  tty (NULL),termFd (NULL),terminal (NULL),
  scenarioTop (scenarioStart),scenarioLength (scenarioLength),
  intVar (intVar),oldPointer(NULL),diffLineOffset (0),
  offset (0),isBreakMode (isBreakMode)
{
  if (maslTty!=NULL)
    {
      tty=new char[strlen (maslTty)+1];
      strcpy (tty,maslTty);
    }
  //do nothing
}

void DebugMode::initializeScreen ()
{
  if (tty!=NULL)
    {
      termFd=fopen (tty,"r+");
      if (termFd==NULL)
	{//terminalopenʤä
	  perror ("can not open the terminal");
	  delete[] tty;
	  tty=NULL;
	}
      else
	{
	  terminal=newterm (NULL,termFd,termFd);
	  set_term (terminal);
	}
  }

  if (tty==NULL)
    {
      if (!(initscr ()))//
	{
	  cerr << "initscr () failed." << endl;
	  exit (-1);//˽
	}
    }


  if (LINES<MIN_LINES||COLS<MIN_COLS)//οŬ
    {//̤롣
      endwin ();
      cerr << "screen is too small." << endl;
      cerr << "screen size must be larger than "
	   << MIN_LINES << "x" << MIN_COLS << "." << endl;
      exit (-1);
    }

  ///////////////////////////////////
  initializeWindow (stdscr);

  /////////////////////////////////////windowγݵڤӽ
  paramWindowLines=PARAMWINDOW_LINES;
  paramWindowCols=COLS;
  paramWindow=newwin (paramWindowLines,paramWindowCols,0,0);
  initializeWindow (paramWindow);
  printParam ();
  
  varWindowLines=VARWINDOW_LINES;
  varWindowCols=COLS;
  varWindow=newwin (varWindowLines,varWindowCols,LINES-varWindowLines,0);
  initializeWindow (varWindow);
  printVar ();
  
  scenarioWindowLines=LINES-varWindowLines-paramWindowLines;
  scenarioWindowCols=COLS;
  scenarioWindow=newwin (scenarioWindowLines,scenarioWindowCols,
			 paramWindowLines,0);
  initializeWindow (scenarioWindow);
  printScenario (scenarioTop);

  cbreak ();
  noecho ();

  refresh ();
}

void DebugMode::initializeWindow (WINDOW* window)
{
  nodelay (window,true);
  keypad (window,true);
  meta (window,true);
  intrflush (window,true);
}

void DebugMode::endScreen ()
{//ncursesθҤ
  delwin (scenarioWindow);
  delwin (varWindow);
  delwin (paramWindow);
  endwin ();
#ifndef __FreeBSD__
  if (tty!=NULL)
    delscreen (terminal);
#endif
}

DebugMode::~DebugMode ()
{
  endScreen ();
  if (termFd!=NULL) fclose (termFd);
  delete[] tty;
}

void DebugMode::windowSizeChange (const char* const pt)
{
  diffLineOffset=0;
  endScreen ();
  initializeScreen ();
  erase ();
  redrawwin (stdscr);
  //ʤeraseredrawʤȲ̤˥ߤĤ롣
  //Ǥ⤿ޤ˻Ĥͭ롢Ρ
  printParam ();
  printScenario (pt,0);
  printVar ();
  refresh ();
}


void DebugMode::printData(const char* const pt)
{
  printParam ();
  printVar ();
  printScenario (pt);
  refresh ();

  int lineOffset=0;
  for (;;)
    {
      //switch (wgetch (varWindow))
      switch (getch ())
	{
	case 'l'://ľ
	case 'f'://ľ
	case CTRL ('f')://ľ
	case '6'://ľ
	case KEY_RIGHT:
	  offset++;
	  if (offset>=MAXVAR) offset=MAXVAR-1;
	  printParam ();
	  printVar ();
	  refresh ();
	  if (!isBreakMode) return;//Ԥɬפ̵Τǽ
	  break;

	case 'h'://ľ
	case 'b'://ľ
	case CTRL ('b')://ľ
	case '4'://ľ
	case KEY_LEFT:
	  offset--;
	  if (offset<0) offset=0;	  
	  printParam ();
	  printVar ();
	  refresh ();
	  if (!isBreakMode) return;//Ԥɬפ̵Τǽ
	  break;

	case 'c'://ľ
	case 'm'://ľ
	case '\r'://ľ
	case KEY_ENTER:
	  isBreakMode=!(isBreakMode);
	  if (!isBreakMode) return;//Ԥɬפ̵Τǽ
	  break;
	  
	case 's'://ľ
	case ' ':
	  return;//spaceenter1commandʤ
	  break;//ǰΰ
      
	case 'k'://ľ
	case 'p'://ľ
	case CTRL ('p')://ľ
	case 'u'://ľ
	case '8'://ľ
	case KEY_UP:
	  lineOffset++;
	  printParam ();
	  printScenario (pt,lineOffset);
	  refresh ();
	  break;

	case 'V'://ľ
	case 'K'://ľ
	case 'P'://ľ
	case 'z'://ľ
	//case CTRL ('z')://ľ
	case 'U'://ľ
	case KEY_BACKSPACE://ľ
	case KEY_PPAGE://ľ
	case KEY_A3:
	  lineOffset+=scenarioWindowLines/2;
	  printParam ();
	  printScenario (pt,lineOffset);
	  refresh ();
	  break;

	case 'j'://ľ
	case 'n'://ľ
	case CTRL ('n')://ľ
	case '2'://ľ
	case 'd'://ľ
	case KEY_DOWN:
	  lineOffset--;
	  printParam ();
	  printScenario (pt,lineOffset);
	  refresh ();
	  break;
	  
	case 'v'://ľ
	case CTRL ('v')://ľ
	case 'J'://ľ
	case 'N'://ľ
	case 'D'://ľ
	case KEY_NPAGE://ľ
	case KEY_C3:
	  lineOffset-=scenarioWindowLines/2;
	  printParam ();
	  printScenario (pt,lineOffset);
	  refresh ();
	  break;
	  
	case '5'://ľ
	case KEY_B2://ľ
	case KEY_HOME:
	  lineOffset=0;
	  printParam ();
	  printScenario (pt,lineOffset);
	  refresh ();
	  break;

	case CTRL ('l'):
	  windowSizeChange (pt);
	  break;

	case ERR://ľ
	default:
	  if (!isBreakMode) return;//Ԥɬפ̵Τǽ
	  printParam ();
	  refresh ();
	  break;//⤷ʤ
	}
    }
}

#define MAX_FORM_LENGTH 256
void DebugMode::printVar ()
{
  werase (varWindow);
  whline (varWindow,0,varWindowCols);
  wmove (varWindow,1,0);
  char buff[MAX_FORM_LENGTH];
  int x;
  int y=1;

  for (int i=offset;i<MAXVAR;i++)
    {
      sprintf (buff,"@%d=%d ",i,intVar[i]);
      getyx (varWindow,y,x);
      if (varWindowCols<x+(int) strlen (buff))
	{//üޤãΤǲ
	  y++;
	  if (y>=varWindowLines) break;//⤦ԤƤɽǤʤΤǽ
	  waddstr (varWindow,"\n");
	}
      waddstr (varWindow,buff);
    }
  wnoutrefresh (varWindow);
}

void DebugMode::printScenario (const char* const nowPtr,
			       const int lineOffset=0)
{
  ///¹ԤƤԤɽ֤׻
  if (oldPointer==NULL) oldPointer=nowPtr;//ǽ˸ƤФ줿
  diffLineOffset+=countBetweenLines (oldPointer,nowPtr);
  oldPointer=nowPtr;
  if (scenarioWindowLines/3+scenarioWindowLines*2/3/3<diffLineOffset ||
      diffLineOffset<-scenarioWindowLines/3)
    {//̤Ϥ߽ФΤǡ̤ư롣
      diffLineOffset=0;
    }
  const int execLine=scenarioWindowLines/3+lineOffset+diffLineOffset;

  werase (scenarioWindow);
  ////////////////////¹ԤԤ
  if (0<=execLine && execLine<scenarioWindowLines)
    {//줫񤳤ȤƤԤwindowͭ
      wattron (scenarioWindow,A_BOLD);
      mvwaddnstr (scenarioWindow,execLine,LEFT_PADDING,
		  searchBeginningOfLine (nowPtr,0),
		  getBeforeLength (nowPtr));
      waddstr (scenarioWindow,arrow);
      waddnstr (scenarioWindow,nowPtr,getAfterLength (nowPtr));
      wattroff (scenarioWindow,A_BOLD);
    }
  /////////////////¹ԤƤԤΤ
  for (int y=-1;y>-execLine-2;y--)
    {
      if (searchBeginningOfLine (nowPtr,y)<=scenarioTop)
	break;//scenarioƬޤǽ񤤤Τǽ
      if (0<=execLine+y && execLine+y<scenarioWindowLines)
	{//줫񤳤ȤƤԤwindowͭ
	  mvwaddnstr (scenarioWindow,execLine+y,LEFT_PADDING,
		      searchBeginningOfLine (nowPtr,y),
		      getAfterLength (searchBeginningOfLine (nowPtr,y)));
	}
    }
  /////////////////¹ԤƤԤ
  for (int y=1;y<scenarioWindowLines-execLine;y++)
    {
      if (searchBeginningOfLine (nowPtr,y+1)>scenarioTop+scenarioLength)
	break;//scenarioκǸޤ褿齪
      if (0<=execLine+y && execLine+y<scenarioWindowLines)
	{//줫񤳤ȤƤԤwindowͭ
	  mvwaddnstr (scenarioWindow,execLine+y,LEFT_PADDING,
		      searchBeginningOfLine (nowPtr,y),
		      getAfterLength (searchBeginningOfLine (nowPtr,y)));
	}
    }

  wnoutrefresh (scenarioWindow);
}

const char* DebugMode::searchBeginningOfLine
(const char* const pt,const int n) const
{
  int counter=0;
  if (n>0)
    {//n>0ʤΤǿʤࡣ
      for (int i=0;i<n;i++)
	{
	  //if (i!=0)counter++;//ΤǤϤưʤ
	  if (*(pt+counter-2)=='\r' || *(pt+counter-1)=='\n') 
	    counter++;//ƬͭäΤǰĿʤ


	  while (*(pt+counter-2)!='\r' && *(pt+counter-1)!='\n')
	    counter++;//ǼιԤιƬޤ
	}
    }
  else
    {//n<=0ʤΤ
      for (int i=0;i<-n+1;i++)
	{
	  if (i!=0)counter--;	  
	  while (*(pt+counter-2)!='\r' && *(pt+counter-1)!='\n')
	    //Ԥΰ֤褿
	    counter--;//ŪιԤľβԤ˰ܤ
	}
    }
  
  return pt+counter;
}

int DebugMode::getAfterLength (const char* const pt) const
{
  if (*pt=='\r' || *pt=='\n')
    return 0;//Ԥΰ֤pointerͭä

  int counter=0;
  while (*(pt+counter+1)!='\r' && *(pt+counter+2)!='\n')
    counter++;//Ԥνޤǻ
  return counter+1;
}

int DebugMode::getBeforeLength (const char* const pt) const
{
  int counter=0;
  while (*(pt-counter-2)!='\r' && *(pt-counter-1)!='\n')
    counter++;
  return counter;
}

int DebugMode::countBetweenLines
(const char* const beforePointer,const char* const nowPointer) const
{//beforePointernowPointerδ֤ˤԤο롣
  //beforePonter<nowPoiterΤȤο֤
  
  int flag=1;
  const char* tmp=beforePointer;
  const char* end=nowPointer;
  if (nowPointer<beforePointer)
    {
      flag=-1;
      tmp=nowPointer;
      end=beforePointer;
    }

  int lineNumber=0;
  while (tmp<end)
    {
      if (*(tmp-1)=='\r' && *tmp=='\n')
	lineNumber++;

      tmp++;
    }
  
  return flag*lineNumber;
}
#endif //DEBUG_MODE
